/* This file is part of Cloudy and is copyright (C)1978-2007 by Gary J. Ferland
 * For conditions of distribution and use see copyright notice in license.txt */
/*cdInit routine to initialize variables, called at start of calculation */
/*cdEXIT exit handler to terminate calculation, called at end of calculation */
#include "cddefines.h"
#include "cddrive.h"
#include "assertresults.h" 
#include "called.h"
#include "version.h"
#include "input.h"
#include "parse.h"
#include "atmdat.h"
#include "zero.h"
#include "date.h"
#include "optimize.h"
#include "grid.h"
#include "path.h"
/* include MPI header if MPI true 
 * This flag indicates whether we are multi-processing.
 * if this is true then all output happens at the end,
 * if false then each set of results is printed and flushed when it happens.
 * must be set true on parallel machines, false on serial */
#ifdef MPI
#	include <mpi.h>
#endif

static char chMonth[12][4] ={"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec" };
static char chMonthInt[12][3] ={"01","02","03","04","05","06","07","08","09","10","11","12" };

#include <stdio.h>

/*************************************************************************
 *
 * cdMPI sets flag to tell exit handler to call MPI_Finalize 
 *
 ************************************************************************/

/* flag set true when cdMPI called, tells exit handler to clean up MPI */
/* lint error saying never used for anything is fine -- 
 * only is used if MPI flags are set,
 * this is so that we do not call MPI_Finalize when on serial queue on
 * parallel machine - that causes machine to drop core */
static bool lgMPI;

/*cdMPI sets flag telling exit handler to call MPI_Finalize, must
 * call after cdInit when MPI is used */
void cdMPI(void) /* set flag so that exit handler will clean up MPI */
{
	lgMPI = true;
	return;
}

/* =================================================================== */
void cdInit(void)
{
	long i;
	double vtest;
	static bool lgDateSet=false;

	/* set ioQQQ to standard output */
	ioQQQ = stdout;

	/* set ioStdin to standard input, will get line images from here */
	ioStdin = stdin;

	/* set flag saying that we have been called */
	lgcdInitCalled = true;

	/* set ioPrnErr to stderr */
	ioPrnErr = stderr;

	/* but don't want to usually do this */
	lgPrnErr = false;

	/* set MPI flag false, set true with cdMPI()*/
	lgMPI = false;

	/* pick up the path from the environment, if set by user */
	char *path = getenv( "CLOUDY_DATA_PATH" );
	if( path != NULL )
	{
		lgDataPathSet = true;

		/* make sure path fits in our chDataPath array */
		size_t len = strlen(path);
		if( len < sizeof(chDataPath) )
			strcpy( chDataPath , path );
		else
		{
			fprintf( ioQQQ, " User-defined path doesn't fit in chDataPath\n" );
			puts( "[Stop in cdInit]" );
			cdEXIT(EXIT_FAILURE);
		}

		/* set pointer to last valid char */
		char *chEnd = chDataPath + len - 1;

		/* now see what this char is */
		if( *chEnd != '/' && *chEnd != ']' && *chEnd != '\\' )
		{
			/* string is not ended properly, we will have to do it */

			/* first determine if there is room to append character */
			if( ++len >= sizeof(chDataPath) )
			{
				fprintf( ioQQQ, " No room to append in chDataPath\n" );
				puts( "[Stop in cdInit]" );
				cdEXIT(EXIT_FAILURE);
			}

			/* find which character to end with since none set above *
			 * there must have been something within the string to indicate 
			 * what type of system we are on */
			if( strchr( chDataPath, '/' ) != NULL )
			{
				/* there was a forward slash somewhere, so cat another to the end */
				strcat( chDataPath , "/" );
			}
			else if( strchr( chDataPath, '\\' ) != NULL )
			{
				/* there was a backward slash somewhere, so cat another to the end */
				strcat( chDataPath, "\\" );
			}
			else if( strchr( chDataPath, '[' ) != NULL )
			{
				/* there was a [ somewhere, so cat ] to the end */
				/* this should work on a VMS machine, but I have not tried it */
				strcat( chDataPath, "]" );
			}
			/* if none of the above were triggered then we did not find anything, 
			 * just leave it alone, either user knew something we didn't, or will
			 * fail when we try to open the file later on */
		}
	}

	/* set the date saved in date.h into version.chDate if it has not already been sone */
	if( !lgDateSet )
	{
		/* this is the decade part of the year counted after 2000 */
		int nDecade = (YEAR-100)/10;
		char chDecade[2];

		/* create space for these strings */
		version.chDate = (char *)MALLOC( sizeof(char)*INPUT_LINE_LENGTH );
		version.chVersion = (char *)MALLOC( sizeof(char)*INPUT_LINE_LENGTH );

		lgDateSet = true;
		sprintf(chDecade,"%i", nDecade );
		/* set date and version - this code will need to be modified in the year 2110 */
		if( nDecade > 9 )
		{
			fprintf(ioQQQ,"must update formation of chDate in cdInit \n");
			TotalInsanity();
		}
		if( DAY < 10 )
		{
			sprintf( version.chDate , "%s%1i%s0%1i", 
				chDecade,
				YEAR-100-nDecade*10,
				chMonth[MONTH],
				DAY );
			sprintf( version.chVersion , "%s%1i.%s.0%1i", 
				chDecade,
				YEAR-100-nDecade*10,
				chMonthInt[MONTH],
				DAY );
		}
		else
		{
			sprintf( version.chDate , "%s%1i%s%2i", 
				chDecade,
				YEAR-100-nDecade*10,
				chMonth[MONTH],
				DAY );
			sprintf( version.chVersion , "%s%1i.%s.%2i", 
				chDecade,
				YEAR-100-nDecade*10,
				chMonthInt[MONTH],
				DAY );
		}

		version.chCitation = CITATION;
		version.chCitationLatex = CITATION_LATEX;
		version.chCitationShort = CITATION_SHORT;

		/* is this a release version? */
		version.lgRelease = true;

		/* is this a beta version?  0 for no */
		version.nBetaVer = 0;

		/* optional letter could come here for updates to branched version */
		/*strcat( version.chVersion , "a" );*/
		/* change following from * to letter when letters are used */
		strcat( version.chDate , "*" );

		char mode[6];
		if( sizeof(long) == 4 )
			strncpy( mode, "ILP32", 6 );
		else if( sizeof(long) == 8 )
			strncpy( mode, "LP64", 6 );
		else
			strncpy( mode, "?????", 6 );

		/* now generate info on how we were compiled */
		sprintf(version.chInfo , 
			"cdInit compiled on %s in OS %s using the %s %i compiler in %s mode." ,
			__DATE__  ,__OS , __COMP , __COMP_VER, mode);
	}

	/*strcpy( version.chDate, "04Oct103" );
	 * >>chng 05 jan 01, version is now the date in format yy.mm.dd 
	strcpy( version.chVersion, " 96.01 " );*/

	/* initialize lots of constants */
	cdDefines();

	/* >>chng 00 mar 10, moved call for grain block data init to Cloudy, PvH */

	/*test if the following integer types have the correct width*/
	if( sizeof(int16) != 2 || sizeof(uint16) != 2 || sizeof(int32) != 4 || sizeof(uint32) != 4 )
		TotalInsanity();

	/*********************************************************
	 *  on a VAX compile with /G_FLOATING option on FORTRAN; *
	 *  following makes sure this happened.                  *
	 *********************************************************/
	vtest = 1e-35;
	vtest /= 1e35;
	if( vtest == 0. )
	{
		fprintf( ioQQQ, " Something is wrong with the double precision.  Use /g_floating on a VAX\n" );
	}

	/* >>chng 02 feb 27, include this test, provided by Peter van Hoof */
	/* check that we ARE NOT being compiled by g++ version 2.96 */
	/* >>chng 07 jan 04 rjrw, just in case someone else ever ships a
		 broken compiler ;-) */
#	if BROKEN_COMPILER
	fprintf(ioQQQ,"HELP!!! You compiled me with %s %d!\n",__COMP,__COMP_VER);
	fprintf(ioQQQ,"This version of %s cannot compile Cloudy and must not be used!\n",__COMP);
	fprintf(ioQQQ,"Please update %s to a functional version.\n",__COMP);
	puts( "[Stop in cdInit]" );
	cdEXIT(EXIT_FAILURE);
#	endif


	/* initialize some variables dealing with cloudy's interaction with machine environment */
	/* if TALK is true then do standard printout
	 * if false then never say anything */
	called.lgTalk = true;
	/* this flag is needed to turn print on to have effect */
	called.lgTalkIsOK = true;
	/* means talk not forced off by call to cdTalk*/
	called.lgTalkForcedOff = false;

	optimize.lgNoVary = false;
	optimize.lgVaryOn = false;
	optimize.lgOptimr = false;
	grid.lgGrid = false;

	for( i=0; i<NUM_OUTPUT_TYPES; i++ )
	{
		grid.lgOutputTypeOn[i] = false;
	}

	for( i=0; i<LIMPAR; i++ )
	{
		grid.numParamValues[i] = 0;
	}

	/* this is a global variable in assertresults.h, and can be checked by
	 * other routines to see if asserts are ok - (most calculations will not use asserts,
	 * and this will be the only place values are set, although they will be checked in maincl) */
	lgAssertsOK = true;
	lgBigBotch = false;
	lgPrtSciNot = false;

	/* number of lines entered with cdLine
	 * both check that number less than NKRD, the limit
	 * the line save array is defined from 0 through input.nSave */
	input.nSave = -1;

	/* nRead is the number of the command in the input stream - many optimize options
	 * point to it to refer to the original command.  it is incremented before
	 * it is used, so will become 0.  it is the array element within the stack
	 * of emission lines */
	input.nRead = -1;

	/* this is number of init lines read in */
	input.nSaveIni = 0;
	input.lgUnderscoreFound = false;
	input.lgBracketFound = false;

	/* this is sanity check that lines are read in ok */
	for( i=0; i < NKRD; i++ )
	{
		strcpy( input.chCardSav[i], "error! - no line image input" );
	}

	/* start the timer to log execution time */
	cdSetExecTime();

	/* zero out lots of variables */
	zero();

	DEBUG_EXIT( "cdInit()" );
	return;
}


/* =================================================================== */
/*cdEXIT the routine that should be called to exit cloudy */
NORETURN void cdEXIT(int iexit)
{
	/*@-redef@*/
	enum {DEBUG_LOC=false};
	/*@+redef@*/
	if( DEBUG_LOC )
	{
		fprintf(ioQQQ," cdEXIT called\n");
	}

	/* close any open units */
	/* flush ioQQQ */
	fflush( ioQQQ );

	ClosePunchFiles();

	/* this routine must be called upon exit or machine will appear to "hang",
	 * but must not be called on MPI machine when used in serial mode,
	 * or it will drop core */
	if( lgMPI )
	{
		fprintf(ioQQQ," MPI_Finalize called.\n");
#		ifdef MPI
		MPI_Finalize();
#		endif
	}
	exit( iexit );
}


