/* This file is part of Cloudy and is copyright (C)1978-2007 by Gary J. Ferland
 * For conditions of distribution and use see copyright notice in license.txt */
/*ion_recom_calculate calculate radiative and dielectronic recombination rate coefficients */
/*Badnell_rec_init This code is written by Terry Yun, 2005 *
 * It reads rate coefficient fits into 3D arrays and output array.out for testing *
 * The testing can be commented out */
/*Badnell_DR_rate_eval This code is written by Terry Yun, 2005 *
 * It interpolates the rate coefficients in a given temperature.*
   It receives ATOMIC_NUM_BIG, NELECTRONS values, temperature and returns the rate coefficient*
   It returns
        '-2': initial <= final
              init < 0 or init >302 or final < 0 or final > 302
        '-1': the transition is not defined
        '99': unknown invalid entries                         */ 
#include "cddefines.h"
#include "phycon.h"
#include "path.h"
#include "elementnames.h"
#include "atmdat.h"
#include "ionbal.h"
#include "dense.h"

static const int MAX_FIT_PAR_DR = 9;
/* one larger than largest atomic number in Nigel's data file */
static const int ATOMIC_NUM_BIG = 55;
/* number of electrons */
static const int NELECTRONS = 12;
static double C[ATOMIC_NUM_BIG][NELECTRONS][MAX_FIT_PAR_DR]; 
static double E[ATOMIC_NUM_BIG][NELECTRONS][MAX_FIT_PAR_DR];    
static int defn_C[ATOMIC_NUM_BIG][NELECTRONS];
static int defn_E[ATOMIC_NUM_BIG][NELECTRONS];
static int nn_counter[ATOMIC_NUM_BIG][NELECTRONS];

static const int MAX_FIT_PAR_RR = 6;
static double par[ATOMIC_NUM_BIG][NELECTRONS][MAX_FIT_PAR_RR]; 
static int defn[ATOMIC_NUM_BIG][NELECTRONS];

/* #define PRINT_DR */
/* #define PRINT_RR */

#if defined(PRINT_DR) || defined(PRINT_RR)
static const char FILE_NAME_OUT[] = "array.out";
#endif

/*ion_recom_calculate calculate radiative and dielectronic recombination rate coefficients */
void ion_recom_calculate( void )
{
	static double TeUsed = -1;
	long int ion , 
		nelem ,
		i;

	DEBUG_ENTRY( "ion_recom_calculate()" );

	if( fabs(phycon.te/TeUsed - 1. ) > 0.001 )
	{
		/*>>chng 06 jan 31 save mean rec coefficients */
		long int nsumrec[LIMELM];
		for( ion=0; ion<LIMELM; ++ion )
		{
			ionbal.DR_Badnell_rate_coef_mean_ion[ion] = 0.;
			nsumrec[ion] = 0;
		}
		TeUsed = phycon.te;
		/* NB - for following loop important to go over all elements and all
		* ions, not just active ones, since mean DR is used as the guess for
		* DR rates for unknown species.  */
		for( nelem=ipHYDROGEN; nelem<LIMELM; ++nelem )
		{
			for( ion=0; ion<nelem+1; ++ion )
			{
				long int n_bnd_elec_before_recom ,
					n_bnd_elec_after_recom;

				n_bnd_elec_before_recom = nelem-ion;
				n_bnd_elec_after_recom = nelem-ion+1;
#				define DR2SMALL	1e-15
				/* Badnell dielectronic recombination rate coefficients */
				if( (ionbal.DR_Badnell_rate_coef[nelem][ion] = Badnell_DR_rate_eval(
					/* atomic number on physics scale - He - 2 */
					nelem+1, 
					/* number of core electrons before capture of free electron */
					n_bnd_elec_before_recom )) >= 0. )
				{
					ionbal.lgDR_Badnell_rate_coef_exist[nelem][ion] = true;
					if( ionbal.DR_Badnell_rate_coef[nelem][ion] > DR2SMALL)
					{
						/* keep track of mean DR rate for this ion */
						++nsumrec[ion];
						ionbal.DR_Badnell_rate_coef_mean_ion[ion] +=
							log10(ionbal.DR_Badnell_rate_coef[nelem][ion]);
					}
				}
				else
				{
					ionbal.lgDR_Badnell_rate_coef_exist[nelem][ion] = false;
					ionbal.DR_Badnell_rate_coef[nelem][ion] = 0.;
				}
				/* Badnell radiative recombination rate coefficients */
				if( (ionbal.RR_Badnell_rate_coef[nelem][ion] = Badnell_RR_rate_eval(
					/* atomic number on physics scale - He - 2 */
					nelem+1, 
					/* number of core electrons before capture of free electron */
					n_bnd_elec_before_recom )) >= 0. )
				{
					ionbal.lgRR_Badnell_rate_coef_exist[nelem][ion] = true;
				}
				else
				{
					ionbal.lgRR_Badnell_rate_coef_exist[nelem][ion] = false;
					ionbal.RR_Badnell_rate_coef[nelem][ion] = 0.;
				}
				/* save D. Verner's radiative recombination rate coefficient 
				* needed for rec cooling, cm3 s-1 */
				ionbal.RR_Verner_rate_coef[nelem][ion] = 
					t_ADfA::Inst().rad_rec( 
					/* number of bound electrons after recombination */
					nelem+1 , 
					/* number of protons on physics scale */
					n_bnd_elec_after_recom , 
					phycon.te );
			}
		}

		/* this branch starts idiosyncratic single ions */
		/* >>chng 06 jun 21 by Mitchell Martin, added new DR rate coefficient
		 * for Fe XIV to Fe XIII recombination calculation 
		 *>>refer	Fe12	DR	Schmidt E.W. et al. 2006, ApJ, 641, 157L */
		double fitSum = 0;	/*the sum of the fitting parameter calculations*/
		/*fitting parameters used to calculate the DR rate for Fe+13 -> Fe+13*/
		double fe13_c[10]={ 3.55e-4, 2.40e-3, 7.83e-3, 1.10e-2, 3.30e-2, 1.45e-1, 8.50e-2, 2.59e-2, 8.93e-3, 9.80e-3 },
			fe13_E[10]={ 2.19e-2, 1.79e-1, 7.53e-1, 2.21e0, 9.57e0, 3.09e1, 6.37e1, 2.19e2, 1.50e3, 7.86e3 };

		double Fe_Gu_c[9][6] = {
		{ 2.50507e-11, 5.60226e-11, 1.85001e-10, 3.57495e-9, 1.66321e-7, 0. },/*fit params for Fe+6*/
		{ 9.19610e-11, 2.92460e-10, 1.02120e-9, 1.14852e-8, 3.25418e-7, 0. }, /* fitting params for Fe+7 */
		{ 9.02625e-11, 6.22962e-10, 5.77545e-9, 1.78847e-8, 3.40610e-7, 0. }, /* fitting params for Fe+8 */
		{ 9.04286e-12, 9.68148e-10, 4.83636e-9, 2.48159e-8, 3.96815e-7, 0. }, /* fitting params for Fe+9 */
		{ 6.77873e-10, 1.47252e-9, 5.31202e-9, 2.54793e-8, 3.47407e-7, 0. }, /* fitting params for Fe+10 */
		{ 1.29742e-9, 4.10172e-9, 1.23605e-8, 2.33615e-8, 2.97261e-7, 0. }, /* fitting params for Fe+11 */
		{ 8.78027e-10, 2.31680e-9, 3.49333e-9, 1.16927e-8, 8.18537e-8, 1.54740e-7 },/*fit params for Fe+12*/
		{ 2.23178e-10, 1.87313e-9, 2.86171e-9, 1.38575e-8, 1.17803e-7, 1.06251e-7 },/*fit params for Fe+13*/
		{ 2.17263e-10, 7.35929e-10, 2.81276e-9, 1.32411e-8, 1.15761e-7, 4.80389e-8 }/*fit params for Fe+14*/
		},

		Fe_Gu_E[9][6] = {
		{ 8.30501e-2, 8.52897e-1, 3.40225e0, 2.23053e1, 6.80367e1, 0. }, /* fitting params for Fe+6 */
		{ 1.44392e-1, 9.23999e-1, 5.45498e0, 2.04301e1, 7.06112e1, 0. }, /* fitting params for Fe+7 */ 
		{ 5.79132e-2, 1.27852e0, 3.22439e0, 1.79602e1, 6.96277e1, 0. }, /* fitting params for Fe+8 */
		{ 1.02421e-1, 1.79393e0, 4.83226e0, 1.91117e1, 6.80858e1, 0. }, /* fitting params for Fe+9 */
		{ 1.24630e-1, 6.86045e-1, 3.09611e0, 1.44023e1, 6.42820e1, 0. }, /* fitting params for Fe+10 */
		{ 1.34459e-1, 6.63028e-1, 2.61753e0, 1.30392e1, 6.10222e1, 0. }, /* fitting params for Fe+11 */
		{ 7.79748e-2, 5.35522e-1, 1.88407e0, 8.38459e0, 3.38613e1, 7.89706e1 }, /*fitting params for Fe+12*/
		{ 8.83019e-2, 6.12756e-1, 2.36035e0, 9.61736e0, 3.64467e1, 8.72406e1 }, /*fitting params for Fe+13*/
		{ 1.51322e-1, 5.63155e-1, 2.57013e0, 9.08166e0, 3.69528e1, 1.08067e2 } /* fitting params for Fe+14*/
		};

		double s_c[5][2] = {
		{0.1565e-3, 0.1617e-2},	/* fitting parameters for S+1 */
		{0.3026e-3, 0.2076e-1},	/* fitting parameters for S+2 */
		{0.3177e-1, 0.6309e-3}, /* fitting parameters for S+3 */
		{0.2464e-1, 0.5553e-3}, /* fitting parameters for S+4 */
		{0.1924e-1, 0.5111e-3}	/* fitting parameters for s+5 */
		},

		s_E[5][2] = {
		{0.1157e6, 0.1868e6},	/* fitting parameters for S+1 */
		{0.9662e5, 0.1998e6},	/* fitting parameters for S+2 */
		{0.1928e6, 0.6126e5},	/* fitting parameters for S+3 */
		{0.1806e6, 0.3142e5},	/* fitting parameters for S+4 */
		{0.1519e6, 0.4978e5}	/* fitting parameters for s+5 */
		};

		/* this is the temperature in eV evaluated to the 3/2 power */
		double te_eV32 = pow( phycon.te_eV, 1.5);	

		/* do Fe+13 -> Fe+12 from Savin et al. if not already done */
		nelem = ipIRON;
		ion = 12;
		if( !ionbal.lgDR_Badnell_rate_coef_exist[nelem][ion] )
		{
			for( i=0; i<10; i++ )
			{
				fitSum += fe13_c[i] * sexp( fe13_E[i]/phycon.te_eV );
			}

			/* update DR rate is not already done */
			ionbal.lgDR_Badnell_rate_coef_exist[nelem][ion] = true;
			ionbal.DR_Badnell_rate_coef[nelem][ion] = fitSum / phycon.te32;
			++nsumrec[ion];
			ionbal.DR_Badnell_rate_coef_mean_ion[ion] += 
				log10(ionbal.DR_Badnell_rate_coef[nelem][ion]);
		}

		/* >>chng 06 jul 07 by Mitchell Martin, added DR rate coefficient 
		 * calculations for Fe+6->Fe+5 through Fe+14->Fe+13
		 * this is still for nelem = ipIRON from the previous calculation 
		 * starts with Fe+6 -> Fe+5 and does the next ion with each iteration */
		for( ion=0; ion<9; ion++ )
		{
			/* only do this rate if not already done by a previous approximation */
			if( !ionbal.lgDR_Badnell_rate_coef_exist[nelem][ion+5] )
			{
				fitSum = 0; /* resets the fitting parameter calculation */
				for( i=0; i<6; i++ )
				{
					fitSum += Fe_Gu_c[ion][i] * sexp( Fe_Gu_E[ion][i]/phycon.te_eV );
				}
				ionbal.lgDR_Badnell_rate_coef_exist[nelem][ion+5] = true;
				ionbal.DR_Badnell_rate_coef[nelem][ion+5] = fitSum / te_eV32;
				if( ionbal.DR_Badnell_rate_coef[nelem][ion+5] > DR2SMALL )
				{
					++nsumrec[ion+5];
					ionbal.DR_Badnell_rate_coef_mean_ion[ion+5] +=
						log10(ionbal.DR_Badnell_rate_coef[nelem][ion+5]);
				}
			}
		}
		/* this is end of Fe DR rates */

		/* now get mean of the DR rates - may be used for ions with no DR rates */
		for( ion=0; ion<LIMELM; ++ion )
		{
			if( nsumrec[ion] > 0 )
				ionbal.DR_Badnell_rate_coef_mean_ion[ion] = 
				pow(10., ionbal.DR_Badnell_rate_coef_mean_ion[ion]/nsumrec[ion]);
			/*fprintf(ioQQQ,"DEBUG %li %.2e\n", ion , ionbal.DR_Badnell_rate_coef_mean_ion[ion] );*/
		}

		/* >>chng 06 jun 28 by Mitchell Martin, added DR rate coefficient calculations for SII, SIII, SIV, SV, and SVI*/
		nelem = ipSULPHUR;
		/* starts with S+1 -> S0 and does the next ion with each iteration*/
		for( ion=0; ion<5; ion++ )
		{

			/* only do this rate if not already done by a previous approximation
			 * so following used for ion <= 3 */
			if( !ionbal.lgDR_Badnell_rate_coef_exist[nelem][ion] )
			{
				/* >>chng 06 jun 28 by Mitchell Martin, added DR rate 
				 * coefficient  for SII, SIII, SIV, SV, and SVI*/
				fitSum = 0;
				for( i=0; i<2; i++ )
				{
					fitSum += s_c[ion][i] * sexp( s_E[ion][i]/phycon.te );
				}
				ionbal.lgDR_Badnell_rate_coef_exist[nelem][ion] = true;
				ionbal.DR_Badnell_rate_coef[nelem][ion] = fitSum / phycon.te32;
				/*>>chng 07 apr 28, use max of this or mean */
				/* change DR data set for S 
				 * three cases for S DR - ionbal.nDR_S_guess
				 * 0, default larger of guess and Badnell
				 * 1, pure Badnell
				 * 3, scaled oxygen */
				if( ionbal.nDR_S_guess==0 )
				{
					/* default larger of guess or Badnell */
					ionbal.DR_Badnell_rate_coef[nelem][ion] = 
						MAX2( ionbal.DR_Badnell_rate_coef[nelem][ion] ,
						ionbal.DR_Badnell_rate_coef_mean_ion[ion]);
				}
				else if( ionbal.nDR_S_guess==1 )
				{
					/* pure Badnell - compiler will remove this non op */
					ionbal.DR_Badnell_rate_coef[nelem][ion] = 
						ionbal.DR_Badnell_rate_coef[nelem][ion];
				}
				else if( ionbal.nDR_S_guess==2 )
				{
					/* scaled oxygen */
					ionbal.DR_Badnell_rate_coef[nelem][ion] = 
						ionbal.DR_Badnell_rate_coef[ipOXYGEN][ion]*
						ionbal.DR_S_scale[ion];
				}
				else
					TotalInsanity();
			}
		}

		/* this set true with PRINT on any of the Badnell set recombination commands */
		if( ionbal.lgRecom_Badnell_print )
		{
			fprintf(ioQQQ,"DEBUG Badnell recombination RR, then DR, T=%.3e\n", phycon.te );
			for( nelem=ipHYDROGEN; nelem<LIMELM; ++nelem )
			{
				fprintf(ioQQQ,"nelem=%li %s, RR then DR\n",
					nelem , elementnames.chElementNameShort[nelem] );
				for( ion=0; ion<nelem+1; ++ion )
				{
					if( ionbal.lgRR_Badnell_rate_coef_exist[nelem][ion] )
					{
						fprintf(ioQQQ," %.2e", ionbal.RR_Badnell_rate_coef[nelem][ion] );
					}
					else
					{
						fprintf(ioQQQ," %.2e", -1. );
					}
				}
				fprintf(ioQQQ,"\n" );
				for( ion=0; ion<nelem+1; ++ion )
				{
					if( ionbal.lgDR_Badnell_rate_coef_exist[nelem][ion] )
					{
						fprintf(ioQQQ," %.2e", ionbal.DR_Badnell_rate_coef[nelem][ion] );
					}
					else
					{
						fprintf(ioQQQ," %.2e", -1. );
					}
				}
				fprintf(ioQQQ,"\n\n" );
			}
			/* now print mean recombination and standard deviation */
			fprintf(ioQQQ,"mean DR recombination ion mean stddev stddev/mean \n" );
			for( ion=0; ion<LIMELM; ++ion )
			{
				double stddev;
				stddev = 0.;
				for( nelem=ion; nelem<LIMELM; ++nelem )
				{
					stddev += POW2( 
						ionbal.DR_Badnell_rate_coef[nelem][ion]- 
						ionbal.DR_Badnell_rate_coef_mean_ion[ion] );
				}
				stddev = sqrt( stddev / MAX2( 1 , nsumrec[ion]-1 ) );
				fprintf(ioQQQ," %2li %.2e %.2e %.2e\n", 
					ion , 
					ionbal.DR_Badnell_rate_coef_mean_ion[ion] , 
					stddev ,
					stddev / SDIV(ionbal.DR_Badnell_rate_coef_mean_ion[ion]) );
			}
			cdEXIT( EXIT_SUCCESS );
		}
	}

	DEBUG_EXIT( "ion_recom_calculate()" );

	return;
}

/*Badnell_rec_init This code is written by Terry Yun, 2005 *
 * It reads rate coefficient fits into 3D arrays and output array.out for testing *
 * The testing can be commented out */
void Badnell_rec_init( void )
{

	double par_C[MAX_FIT_PAR_DR];
	double par_E[MAX_FIT_PAR_DR];
	char chLine[INPUT_LINE_LENGTH];
	int z_value, n_value;
	int i, j, k;
	int count, number;
	double temp_par[MAX_FIT_PAR_RR];
	int M_state, W_state;

	/*char chFilename[FILENAME_PATH_LENGTH_2];*/
	const int NBLOCK = 2;
	int data_begin_line[NBLOCK];/*it tells you where the data set begins(begins with 'Z')*/
	int length_of_line;	/*this variable checks for a blank line*/
	FILE *ioDATA;
	char chFilename[FILENAME_PATH_LENGTH_2];
	int yr, mo, dy;
	char *chs;

	DEBUG_ENTRY( "Badnell_rec_init()" );

#	if defined(PRINT_DR) || defined(PRINT_RR)
	FILE *ofp;
	ofp= fopen(FILE_NAME_OUT, "w");
	if( ofp == NULL )
	{
		printf("Can't open %s\n", FILE_NAME_OUT);
		exit(1);
	}
#	endif

	for( i=0; i<NBLOCK; ++i )
	{
		/* set to really large negative number - crash if used before being redefined */
		data_begin_line[i] = INT_MIN;
	}

	/* check on path is file not here and path set */
	/* path was parsed in getset */
	if( lgDataPathSet )
	{
		/*path set, so look only there */
		strcpy( chFilename , chDataPath );
		strcat( chFilename , "badnell_dr.dat" );
	}
	else
	{
		/* path not set, check local space only */
		strcpy( chFilename , "badnell_dr.dat" );
	}

	if( ( ioDATA = fopen( chFilename , "r" ) ) == NULL )
	{
		printf("Badnell_rec_init cannot open %s\n", chFilename);
		/* make announcement that path probably not set */
		path_not_set();
		cdEXIT(EXIT_FAILURE);
	}

	count = 0;
	number = 0;

	/*Find out the number line where the data starts 
	 * there are two main blocks of data and each starts with a Z in column 2 */
	while( fgets(chLine, (int)sizeof(chLine), ioDATA) != NULL )
	{
		count++;

		if( chLine[2]=='Z' )
		{
			/* number has to be 0 or 1, and indicates the first or second block of data
			 * count is the line number for the start of that block */
			data_begin_line[number] = count;
			ASSERT( number < NBLOCK );
			number++;
		}
	}

	/*set a flag for a undefined data*/
	for( i=0; i<ATOMIC_NUM_BIG; i++ )
	{
		for( j=0; j<NELECTRONS; j++ )
		{
			defn_C[i][j] = 0;
			defn_E[i][j] = 0;   
		}
	}

	/*set fitting coefficients to zero initially*/
	for( i=0; i<ATOMIC_NUM_BIG; i++ )
	{
		for( j=0; j<NELECTRONS; j++ )
		{
			for( k=0; k<MAX_FIT_PAR_DR; k++ )
			{
				C[i][j][k] = 0;
				E[i][j][k] = 0;  
			}
		}
	}

	count = 0;
	/*Start from beginning to read in again*/  
	fseek(ioDATA, 0, SEEK_SET);
	/* read magic number for DR data */
	fgets(chLine, (int)sizeof(chLine), ioDATA);
	count++;
	{
		int dr_yr = 2006, dr_mo = 3, dr_dy = 11;
		chs = strchr(chLine, ')');
		++chs;
		sscanf(chs, "%4i%2i%2i",&yr, &mo, &dy);
		if((yr != dr_yr) || (mo != dr_mo) || (dy != dr_dy))
		{
			printf("The version of %s I found (%i %i %i) is not the current version (%i %i %i).\n",
			       chFilename, yr, mo, dy, dr_yr, dr_mo, dr_dy);
			cdEXIT(EXIT_FAILURE);
		}
	}

	while( fgets(chLine, (int)sizeof(chLine), ioDATA) != NULL )
	{
		count++;
		length_of_line = (int)strlen(chLine);

		/*reading in coefficients C */
		if( count > data_begin_line[0] && count < data_begin_line[1] && length_of_line >3 )
		{
			/*set array par_C to zero */
			for( i=0; i<MAX_FIT_PAR_DR; i++ )
			{
				par_C[i] = 0;
			}
			sscanf(chLine, "%i%i%i%i%lf%lf%lf%lf%lf%lf%lf%lf%lf",
			       &z_value, &n_value, &M_state, &W_state, &par_C[0], &par_C[1], &par_C[2],
			       &par_C[3], &par_C[4], &par_C[5], &par_C[6], &par_C[7], &par_C[8]);

			if(M_state == 1)
			{
				/*Set a flag to '1' when the indices are defined */
				defn_C[z_value][n_value] = 1;

				/*counting the number of coeffs */
				nn_counter[z_value][n_value] = 8;   
				while( par_C[nn_counter[z_value][n_value]] == 0 )
				{
					--nn_counter[z_value][n_value];
					if( nn_counter[z_value][n_value] < 0 )
						break;
				}
				++nn_counter[z_value][n_value];

			/*assign the values into array */
				for( i=0; i<nn_counter[z_value][n_value]; i++ )
					C[z_value][n_value][i] = par_C[i];
			}
		}
	}

	/*starting again to read in E values */
	fseek(ioDATA, 0, SEEK_SET);
	count = 0; 
	while( fgets(chLine, (int)sizeof(chLine), ioDATA) != NULL )
	{
		count++;
		length_of_line = (int)strlen(chLine);
		if( count > data_begin_line[1] && length_of_line > 3 )
		{

			/*set array par_E to zero*/
			for( i=0; i<MAX_FIT_PAR_DR; i++ )
			{
				par_E[i] = 0;
			}
			sscanf(chLine, "%i%i%i%i%lf%lf%lf%lf%lf%lf%lf%lf%lf",
			       &z_value, &n_value, &M_state, &W_state, &par_E[0], &par_E[1], &par_E[2],
			       &par_E[3], &par_E[4], &par_E[5], &par_E[6], &par_E[7], &par_E[8]);

			if(M_state == 1)
			{
				/*counting the number of coeffs*/
				nn_counter[z_value][n_value] = 8;
				while( par_E[nn_counter[z_value][n_value]] == 0 )
				{
					--nn_counter[z_value][n_value];
					
					if( nn_counter[z_value][n_value] < 0 )
						break;
				}
				++nn_counter[z_value][n_value];

				/*Set a flag to '1' when the indices are defined*/
				defn_E[z_value][n_value] = 1;

				/*assign the values into array*/
				for( i=0; i<nn_counter[z_value][n_value]; i++ )
					E[z_value][n_value][i] = par_E[i];
			}
		}
	}

	/*output coefficients for defined values for testing */
#	ifdef PRINT_DR
	for( i=0; i<ATOMIC_NUM_BIG; i++ )
	{
		for( j=0; j<NELECTRONS; j++ )
		{
			if( defn_C[i][j] == 1 )
			{
				fprintf(ofp, "%i %i %e %e %e %e %e %e %e %e %e\n",
					i, j, C[i][j][0], C[i][j][1], C[i][j][2], C[i][j][3], C[i][j][4],
					C[i][j][5], C[i][j][6], C[i][j][7], C[i][j][8]);
			}
		}
	}
	for( i=0; i<ATOMIC_NUM_BIG; i++ )
	{
		for( j=0; j<NELECTRONS; j++ )
		{
			if( defn_E[i][j] == 1 )
			{
				fprintf(ofp, "%i %i %e %e %e %e %e %e %e %e %e\n",
					i, j, E[i][j][0], E[i][j][1], E[i][j][2], E[i][j][3], E[i][j][4],
					E[i][j][5], E[i][j][6], E[i][j][7], E[i][j][8]);
			}
		}
	}
	fclose(ofp);
#	endif

	/*checking for the match of defn_C and defn_E - Both have to be defined*/
	for( i=0; i<ATOMIC_NUM_BIG; i++ )
	{
		for( j=0; j<NELECTRONS; j++ )
		{
			if( defn_C[i][j] != defn_E[i][j] )
			{
				printf("C%i, E%i: %i %i\n", i, j, defn_C[i][j], defn_E[i][j]);
				printf("Critical error in defn\n");
			}
		}
	}

	/* fclose(ioDATA); */

	DEBUG_EXIT( "Badnell_rec_init()" );

	/***************************/
	/* addition: code for RR(Radiative recombination) */

	if( lgDataPathSet )
	{
		/*path set, so look only there */
		strcpy( chFilename , chDataPath );
		strcat( chFilename , "badnell_rr.dat" );
	}
	else
	{
		/* path not set, check local space only */
		strcpy( chFilename , "badnell_rr.dat" );
	}

	if( ( ioDATA = fopen( chFilename , "r" ) ) == NULL )
	{
		printf("Badnell_rec_init cannot open %s\n", chFilename);
		exit(1);
	}

	/*set a flag for a undefined data*/
	for( i=0; i<ATOMIC_NUM_BIG; i++ )
	{
		for( j=0; j<NELECTRONS; j++ )
		{
			defn[i][j] = 0;
		}
	}

	/*set fitting coefficients to zero initially*/
	for( i=0; i<ATOMIC_NUM_BIG; i++ )
	{
		for( j=0; j<NELECTRONS; j++ )
		{
			for( k=0; k<MAX_FIT_PAR_RR; k++ )
			{
				par[i][j][k] = 0;
			}
		}
	}

	/* read magic number for RR data */
	{
		int rr_yr = 2005, rr_mo = 12, rr_dy = 27;
		fgets(chLine, (int)sizeof(chLine), ioDATA);
		chs = strchr(chLine, ')');
		++chs;
		sscanf(chs, "%4i%2i%2i", &yr, &mo, &dy);
		if((yr != rr_yr)||(mo != rr_mo)||(dy != rr_dy))
		{
			printf("The version of %s I found (%i %i %i) is not the current version (%i %i %i).\n",
			       chFilename, yr, mo, dy, rr_yr, rr_mo, rr_dy);
			cdEXIT(EXIT_FAILURE);
		}
	}

	while( fgets(chLine, (int)sizeof(chLine), ioDATA) != NULL )
	{
		/*reading in coeffs */
		/*set array par to zero */
		for( i=0; i<MAX_FIT_PAR_RR; i++ )
		{
			temp_par[i] = 0;
		}
		if(chLine[0] != '#')
		{
			sscanf(chLine, "%i%i%i%i%lf%lf%lf%lf%lf%lf",
			       &z_value, &n_value, &M_state, &W_state, &temp_par[0], &temp_par[1],
			       &temp_par[2], &temp_par[3], &temp_par[4], &temp_par[5]);

			if(M_state == 1)
			{
				/*Set a flag to '1' when the indices are defined */  
				defn[z_value][n_value] = 1;
				/*assign the values into array */
				for( i=0; i<MAX_FIT_PAR_RR; i++ )
					par[z_value][n_value][i] = temp_par[i];
			}
		}
	}

	/*output coefficients for defined values for testing */
#	ifdef PRINT_RR
	count = 0;
	for( i=0; i<ATOMIC_NUM_BIG; i++ )
	{
		for( j=0; j<NELECTRONS; j++ )
		{
			if( defn[i][j] == 1 )
			{
				fprintf(ofp, "%i %i %e %e %e %e %e %e\n",
					i, j, par[i][j][0], par[i][j][1], par[i][j][2], par[i][j][3],
					par[i][j][4], par[i][j][5]);
				count++;
			}
		}
	}
	fprintf(ofp, "total lines are %i ", count);

	fclose(ofp);
#	endif

	fclose(ioDATA);

	{
		/*@-redef@*/
		enum {DEBUG_LOC=false};
		/*@+redef@*/
		if( DEBUG_LOC )
		{
			long int nelem;
			for( nelem=ipHYDROGEN; nelem<LIMELM; ++nelem )
			{
				int ion;
				fprintf(ioQQQ,"\nDEBUG rr rec\t%li",nelem);
				for( ion=0; ion<=nelem; ++ion )
				{
					fprintf(ioQQQ,"\t%.2e", Badnell_RR_rate_eval(nelem+1 , nelem-ion ) );
				}
				fprintf(ioQQQ,"\n");
				fprintf(ioQQQ,"DEBUG dr rec\t%li",nelem);
				for( ion=0; ion<=nelem; ++ion )
				{
					fprintf(ioQQQ,"\t%.2e", Badnell_DR_rate_eval(nelem+1 , nelem-ion ) );
				}
				fprintf(ioQQQ,"\n");
			}
			cdEXIT(EXIT_FAILURE);
		}
	}

	return;
}

/*Badnell_DR_rate_eval This code is written by Terry Yun, 2005 *
 * It interpolates the rate coefficients in a given temperature.*
   It receives ATOMIC_NUM_BIG, NELECTRONS values, temperature and returns the rate coefficient*
   It returns
        '-2': initial <= final
              init < 0 or init >302 or final < 0 or final > 302
        '-1': the transition is not defined
        '99': unknown invalid entries                         */ 
double Badnell_DR_rate_eval(
	/* atomic number on physics scale - He - 2 */
	int n_atomic_number, 
	/* number of core electrons before capture of free electron */
	int n_core_e_before_recomb )
{

	double k, sum;
	int i;

	DEBUG_ENTRY( "Badnell_DR_rate_eval()" );

#	if 0
	if( n_atomic_number == 26 && n_core_e_before_recomb==13 )
	{
		/* >>chng 06 aug 10, add this 
		 * special case, Fe+13 -> Fe+12, DR from
		 * >>refer	Fe13	DR	Badnell, N. 2006, J Phys B in press 
		 * Fe 3p^q, q=1*/
		/*c1 c2 c3 c4 c5 c6 c7 c8*/
		double cFe13[8] =
		{1.090e-3, 7.801e-3, 1.132e-2, 4.740e-2, 1.990e-1, 3.379e-2, 1.140e-1, 1.250e-1};
		/*E1 E2 E3 E4 E5 E6 E7 E8*/
		double EFe13[8] =
		{1.246e3, 1.063e4, 4.719e4, 1.952e5, 5.637e5, 2.248e6, 7.202e6, 3.999e9};

		sum = 0;
		i = 0;
		/* loop over all non-zero terms */
		for(i=0; i<8; ++i  )
		{
			sum += (cFe13[i] * sexp( EFe13[i]/phycon.te));
		}

		/*k = pow(phycon.te, -1.5) * sum;*/
		k = sum / phycon.te32;

		DEBUG_EXIT( "Badnell_DR_rate_eval()" );

		return k;

	}
#	endif
	if( n_atomic_number==26 && n_core_e_before_recomb>=12 && n_core_e_before_recomb<=18 )
	{
		/* these data are from Badnell preprint submitted to ApJ aug 2006, they do
		 * Fe 8+ to Fe 12+, but also include Fe13+ and Fe 14+,
		 * so these are 26-8=18 to 26-14=12 
		 * increasing number of bound electrons, 0 is 14 elec, 1 is 15 elec 
		 * Fe 3p^q, q=2-6*/
		double cFe_q[7][8] =
		{
			{5.636e-4, 7.390e-3, 3.635e-2, 1.693e-1, 3.315e-2, 2.288e-1, 7.316e-2, 0.},
			{1.090e-3, 7.801e-3, 1.132e-2, 4.740e-2, 1.990e-1, 3.379e-2, 1.140e-1, 1.250e-1},
			{3.266e-3, 7.637e-3, 1.005e-2, 2.527e-2, 6.389e-2, 1.564e-1, 0.,       0.},
			{1.074e-3, 6.080e-3, 1.887e-2, 2.540e-2, 7.580e-2, 2.773e-1, 0.,       0.},
			{9.073e-4, 3.777e-3, 1.027e-2, 3.321e-2, 8.529e-2, 2.778e-1, 0.,       0.},
			{5.335e-4, 1.827e-3, 4.851e-3, 2.710e-2, 8.226e-2, 3.147e-1, 0.,       0.},
			{7.421e-4, 2.526e-3, 4.605e-3, 1.489e-2, 5.891e-2, 2.318e-1, 0.,       0.}
		};


		double EFe_q[7][8] =
		{
			{3.628e3, 2.432e4, 1.226e5, 4.351e5, 1.411e6, 6.589e6, 1.030e7, 0},
			{1.246e3, 1.063e4, 4.719e4, 1.952e5, 5.637e5, 2.248e6, 7.202e6, 3.999e9},
			{1.242e3, 1.001e4, 4.466e4, 1.497e5, 3.919e5, 6.853e5, 0.     , 0.},
			{1.387e3, 1.048e4, 3.955e4, 1.461e5, 4.010e5, 7.208e5, 0.     , 0.},
			{1.525e3, 1.071e4, 4.033e4, 1.564e5, 4.196e5, 7.580e5, 0.     , 0.},
			{2.032e3, 1.018e4, 4.638e4, 1.698e5, 4.499e5, 7.880e5, 0.     , 0.},
			{3.468e3, 1.353e4, 3.690e4, 1.957e5, 4.630e5, 8.202e5, 0.     , 0.}
		};
		/* nion is for the above block of numers */
		long int nion = n_core_e_before_recomb - 12;
		ASSERT( nion>=0 && nion <=6 );

		sum = 0;
		i = 0;
		/* loop over all non-zero terms */
		for(i=0; i<8; ++i  )
		{
			sum += (cFe_q[nion][i] * sexp( EFe_q[nion][i]/phycon.te));
		}

		/*k = pow(phycon.te, -1.5) * sum;*/
		k = sum / phycon.te32;

		DEBUG_EXIT( "Badnell_DR_rate_eval()" );

		return k;


	}
	/*Invalid entries returns '-2':if the z_values are smaller than equal to the n_values */
	else if( n_atomic_number <= n_core_e_before_recomb )
	{
		k = -2;
	}
	/*Invalid entries returns '-2' if n_atomic_number and n_core_e_before_recomb are out of the range*/
	else if( n_atomic_number < 2 || n_atomic_number > 30 ||
		 n_core_e_before_recomb < 1 || n_core_e_before_recomb > 11 )
	{
		k = -2;
	}
	/*undefined z and n returns '-1'*/
	else if( defn_C[n_atomic_number][n_core_e_before_recomb] == 0 )
	{
		k = -1;
	}
	else if( defn_C[n_atomic_number][n_core_e_before_recomb] == 1 )
	{
		sum = 0;
		i = 0;
		/* loop over all non-zero terms */
		for(i=0; i<nn_counter[n_atomic_number][n_core_e_before_recomb]; ++i  )
		{
			sum += (C[n_atomic_number][n_core_e_before_recomb][i] *
				sexp( E[n_atomic_number][n_core_e_before_recomb][i]/phycon.te));
		}

		/*k = pow(phycon.te, -1.5) * sum;*/
		k = sum / phycon.te32;
	}
	/*unknown invalid entries returns '-99'*/
	else
	{
		k = -99;
	}

	DEBUG_EXIT( "Badnell_DR_rate_eval()" );

	return k;

}

double Badnell_RR_rate_eval(
	/* atomic number on physics scale - He - 2 */
	int n_atomic_number, 
	/* number of core electrons before capture of free electron */
	int n_core_e_before_recomb )
{
	double k;
	double B, D, F;

	DEBUG_ENTRY( "Badnell_DR_rate_eval()" );

#	if 0
	if( n_atomic_number == 26 && n_core_e_before_recomb==13 )
	{
		/* >>chng 06 aug 10, add this 
		 * special case, Fe+13 -> Fe+12, DR from
		 * >>refer	Fe13	DR	Badnell, N. 2006, J Phys B in press */
		double parFe13[6] = 
		{4.321e-10, 0.6091, 2.255e03, 4.962e07, 0.0356, 1.006e05};
		/* k=A*[(T/T0)^1/2*(1+(T/T0)^1/2)^1-B*(1+(T/T1)^1/2)^1+B]^-1 
		where B = B + C*exp(-T2/T) */
		double temp;

		temp = -parFe13[5]/phycon.te; /* temp = (-T2/T) */
		B = parFe13[1] + parFe13[4]*exp(temp);
		D = sqrt(phycon.te/parFe13[2]); /* D = (T/T0)^1/2 */
		F = sqrt(phycon.te/parFe13[3]); /* F = (T/T1)^1/2 */
		k = parFe13[0]/(D*pow((1.+D),(1.-B))*pow((1.+F),(1.+B)));

		DEBUG_EXIT( "Badnell_RR_rate_eval()" );

		return k;
	}
#	endif
	if( n_atomic_number==26 && n_core_e_before_recomb>=12 && n_core_e_before_recomb<=18 )
	{
		/* >>chng 06 aug 16, this includes previous ions - above block no longer needed */
		double parFeq[7][6] ={
			{1.179e-9 , 0.7096, 4.508e2, 3.393e7, 0.0154, 3.977e6},
			{1.050e-9 , 0.6939, 4.568e2, 3.987e7, 0.0066, 5.451e5},
			{9.832e-10, 0.7146, 3.597e2, 3.808e7, 0.0045, 3.952e5},
			{8.303e-10, 0.7156, 3.531e2, 3.554e7, 0.0132, 2.951e5},
			{1.052e-9 , 0.7370, 1.639e2, 2.924e7, 0.0224, 4.291e5},
			{1.338e-9 , 0.7495, 7.242e1, 2.453e7, 0.0404, 4.199e5},
			{1.263e-9 , 0.7532, 5.209e1, 2.169e7, 0.0421, 2.917e5}
		};

		double temp;
		/* nion is for the above block of numers */
		long int nion = n_core_e_before_recomb - 12;
		ASSERT( nion>=0 && nion <=6 );

		temp = -parFeq[nion][5]/phycon.te; /* temp = (-T2/T) */
		B = parFeq[nion][1] + parFeq[nion][4]*exp(temp);
		D = sqrt(phycon.te/parFeq[nion][2]); /* D = (T/T0)^1/2 */
		F = sqrt(phycon.te/parFeq[nion][3]); /* F = (T/T1)^1/2 */
		k = parFeq[nion][0]/(D*pow((1.+D),(1.-B))*pow((1.+F),(1.+B)));

		DEBUG_EXIT( "Badnell_RR_rate_eval()" );

		return k;
	}


	/*Invalid entries returns '-2':if the z_values are smaller than equal to the n_values */
	else if( n_atomic_number <= n_core_e_before_recomb )
	{
		k = -2;
	}
	/*Invalid entries returns '-2' if n_atomic_number and n_core_e_before_recomb are out of the range*/
	else if( n_atomic_number < 1 || n_atomic_number > 30 ||
		 n_core_e_before_recomb < 0 || n_core_e_before_recomb > 11 )
	{
		k = -2;
	}
	/*undefined z and n returns '-1'*/
	else if( defn[n_atomic_number][n_core_e_before_recomb] == 0 )
	{
		k = -1;
	}
	/* coefficients:A=par[0], B=par[1], T0=par[2], T1=par[3], C=par[4], T2=par[5] */
	else if( defn[n_atomic_number][n_core_e_before_recomb] == 1 )
	{

		/* k=A*[(T/T0)^1/2*(1+(T/T0)^1/2)^1-B*(1+(T/T1)^1/2)^1+B]^-1 
		where B = B + C*exp(-T2/T) */
		double temp;

		temp = -par[n_atomic_number][n_core_e_before_recomb][5]/phycon.te; /* temp = (-T2/T) */
		B = par[n_atomic_number][n_core_e_before_recomb][1] + par[n_atomic_number][n_core_e_before_recomb][4]*exp(temp);
		D = sqrt(phycon.te/par[n_atomic_number][n_core_e_before_recomb][2]); /* D = (T/T0)^1/2 */
		F = sqrt(phycon.te/par[n_atomic_number][n_core_e_before_recomb][3]); /* F = (T/T1)^1/2 */
		k = par[n_atomic_number][n_core_e_before_recomb][0]/(D*pow((1.+D),(1.-B))*pow((1.+F),(1.+B)));

	}

	/*unkown invalid entries returns '-99'*/
	else
	{
		k = -99;
	}

	DEBUG_EXIT( "Badnell_DR_rate_eval()" );
	return k;
}
